#! /usr/bin/lua
-- vim:sw=4:sts=4

require "gtk2"

-- meta class for my window
MYWIN = {}
win_count = 0


function mywin_new()
    local self = { pixmap=nil }
    setmetatable(self, MYWIN)
    self.win = gtk.window_new(gtk.GTK_WINDOW_TOPLEVEL)
    self.win:connect('destroy', MYWIN.on_destroy, self)
    self.win:set_title('Pixmap Test')
    local da = gtk.drawing_area_new()
    self.win:add(da)
    da:connect('configure-event', MYWIN.on_configure, self)
    da:connect('expose-event', MYWIN.on_expose, self)
    self.win:show_all()
    win_count = win_count + 1
    return self
end

function MYWIN:on_destroy()
    win_count = win_count - 1
    if win_count == 0 then
	gtk.main_quit()
    end
end

-- On configure (size change), allocate a pixmap and fill with white.
function MYWIN:on_configure(da, ev)
    print "on_configure"
    local window = self.win.window
    local width, height = window:get_size(0, 0)
    -- deallocate previous pixmap
    if (self.pixmap) then self.pixmap:unref() end
    -- allocates memory in X server...
    self.pixmap = gtk.call("gdk_pixmap_new", window, width, height, -1)
    local style = self.win:get_style()
    local white_gc = style.white_gc
    local black_gc = style.black_gc

    -- clear the whole pixmap
    self.pixmap:draw_rectangle(white_gc, true, 0, 0, width, height)

    -- draw a rectangle
    if width > 20 and height > 20 then
	self.pixmap:draw_rectangle(black_gc, false, 10, 10, width - 20,
	    height - 20)
    end

    -- draw a text message
    local message = "Hello, World!"
    local layout = self.win:create_pango_layout(message)
    self.pixmap:draw_layout(black_gc, 15, 15, layout)

    -- get size of message
    local region = layout:get_clip_region(0, 0, {0, string.len(message)}, 1)
    local rect = gtk.new("GdkRectangle")
    region:get_clipbox(rect)
    if rect.width > 0 then
	self.pixmap:draw_layout(black_gc, width - 15 - rect.width,
	    height - 15 - rect.height, layout)
    end

    -- collectgarbage()
    -- print(gcinfo())

    return true
end

-- on expose, copy the newly visible part of the pixmap to the GdkWindow
-- of the drawing area.
function MYWIN:on_expose(da, ev)
    print "on_expose"
    local area = ev.expose.area
    local x, y, w, h = area.x, area.y, area.width, area.height
    local window = da.window
    local style = self.win:get_style()
    local white_gc = style.white_gc
    gtk.call("gdk_draw_drawable", window, white_gc, self.pixmap,
	x, y, x, y, w, h)
    return false
end

gtk.init(nil, nil)
mywin1 = mywin_new()
mywin2 = mywin_new()
gtk.main()

