#! /usr/bin/perl -w
# Generate a function list to be used by libluagtk2.
# Runs on Linux, but the resulting data file should be usable on other
# platforms, too.
# Copyright (C) 2005 Wolfgang Oertl
#
# Revision history:
#  2005-07-14	extract data types, function list and output.  Structure info
#		started but not yet working.
#
#

use strict;

my $tmpfile = "tmp.$$.c";
my $tmpo = "tmp.$$.o";
my $tmpfile2 = "tmp.$$.l";
my $ofile = "gtkdata";

# For these Lua datatypes also store the exact data type.  These
# types get codes starting at 'a', the others starting at 'A'.
# XXX will go away.
#my %with_detail = (
#	"WIDGET"	=> 1,
#	"WIDGETPTR"	=> 1,
#	"STRUCT"	=> 1,
#	"STRUCTPTR"	=> 1,
#	"STRUCTPTRPTR"	=> 1,
#	"UNION"		=> 1,
##	"ENUM"		=> 1,
##	"ENUMPTR"	=> 1,
##	"ENUMPTRPTR"	=> 1,
##	"FUNC"		=> 1,
##	"FUNCPTR"	=> 1,
#);


open TMP, ">$tmpfile" or die;
print TMP "#define GTK_DISABLE_DEPRECATED 1\n";
print TMP "#include <gtk/gtk.h>\n";
close TMP;

system "cc -I \$(pkg-config gtk+-2.0 --cflags) -gstabs -c -o $tmpo -aux-info $tmpfile2 $tmpfile";
unlink $tmpfile;

# key=name, value=number of times used
my %c_types;

# key=name, value=(return type, arg1, arg2, ...)
my %functions;

# key=name, value=value
my %enum;

# -- read a list of functions with their parameters --

my ($line, $func, $args, @w, $ret_type, $target, $arg);
open LIST, $tmpfile2 or die;
while (<LIST>) {
	chop;
	next unless /:NC \*\/ extern (.*)$/;
	next if /\(\*\)/;
	$line = $1;

	# extract return type + function name and parameters
	$line =~ /^(.*?) \((.*?)\);$/;
	$func = $1;
	$args = $2;

	# split return type and function name
	@w = split(" ", $func);
	$func = pop(@w);
	$ret_type = join(" ", @w);
	while (substr($func, 0, 1) eq "*") {
		$ret_type .= "*";
		$func = substr($func, 1);
	}

	# skip non-gtk functions
	next unless $func =~ /^(pango_|gtk_|gdk_|g_|atk_)/;

	$args = "$ret_type, $args";

	# split parameters
	$args =~ s/const ?//g;
	$args =~ s/ \*/\*/g;
	@w = split(/ ?, /, $args);

	# function without parameters?
	if ($#w == 1 and $w[1] eq "void") {
		pop @w;
	}
	
	# gtk_..._new_ function that returns GtkWidget?
	if ($w[0] eq "GtkWidget*" and $func =~ /^gtk_(.*?)_new/) {
		$target = ucfirst($1);
		$target =~ s/^([VH][a-z])/uc($1)/ge;
		$target =~ s/HAndle/Handle/;
		$target =~ s/VIewport/Viewport/;
		$target =~ s/_([a-z])/uc($1)/ge;
		$target = "Gtk" . $target . "*";
		# print STDERR "remap GtkWidget* of $func to $target\n";
		$w[0] = $target;
	}

	$c_types{$w[0]} ++;

	# other parameters
	for $arg (@w[1 .. $#w]) {
		$c_types{$arg} ++;
	}

	$functions{$func} = [ @w ];
}
close LIST;
unlink $tmpfile2;


# mapping of Gtk (C) types to Lua types
# hierarchy of Gtk Classes

my %typedefs;

# information for all possible types (not all of then are actually used).
my %ffi_types = (
	# name => extradata?, AT_xxx, ffi_type_xxx
	"BOOL"		=> [ 0, "AT_BOOL",	"uint" ],
	"BOOLPTR" 	=> [ 0, "AT_POINTER",	"pointer" ],
	"DOUBLE"	=> [ 0, "AT_DOUBLE",	"double" ],
	"DOUBLEPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"ENUM"		=> [ 0, "AT_LONG",	"uint" ],
	"ENUMPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"ENUMPTRPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"FUNC"		=> [ 0, "AT_POINTER",	"pointer" ],
	"FUNCPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"LONG"		=> [ 0, "AT_LONG",	"uint" ],
	"LONGLONG"	=> [ 0, "AT_LONG",	"ulong" ],
	"LONGPTR"	=> [ 0, "AT_LONGPTR",	"pointer" ],
	"LONGPTRPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"PTR"		=> [ 0, "AT_POINTER",	"pointer" ],
	"PTRPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"STRING"	=> [ 0, "AT_STRING",	"pointer" ],
	"STRINGPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"STRINGPTRPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"STRUCT"	=> [ 1, "AT_STRUCT",	undef ],
	"STRUCTPTR"	=> [ 1, "AT_STRUCTPTR",	"pointer" ],
	"STRUCTPTRPTR"	=> [ 1, "AT_POINTER",	"pointer" ],
	"STRUCTPTRPTRPTR"=>[ 1, "AT_POINTER",	"pointer" ],
	"UNION"		=> [ 1, "AT_LONG",	"uint" ],	# unusual!
	"UNIONPTR"	=> [ 1, "AT_POINTER",	"pointer" ],
	"VALIST"	=> [ 0, "AT_VALIST",	undef ],
	"VARARG"	=> [ 0, "AT_VARARG",	undef ],
	"VOID"		=> [ 0, "AT_VOID",	"void" ],
	"VOIDPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"VOIDPTRPTR"	=> [ 0, "AT_POINTER",	"pointer" ],
	"WIDGET"	=> [ 1, "AT_WIDGET",	"pointer" ],
	"WIDGETPTR"	=> [ 1, "AT_POINTER",	"pointer" ],
);

# predefined typedefs.  All basic types must map to an uppercase "Lua" type
# from the hash %ffi_types.
my %typedefs_override = (
	"GdkAtom" => "PTR",			# simple pointer
	"GtkMenuEntry" => "STRUCT",
	"GdkFilterFunc" => "FUNC",
#	"gpointer" => "PTR",
#	"gconstpointer" => "PTR",
	"gboolean" => "BOOL",
	"char*" => "STRING",
	"void" => "VOID",
	"struct tm" => "STRUCT",
	"fpos_t" => "STRUCT",
	"void *" => "PTR",			# required??

	# some omissions
	"GtkTextLayout" => "struct _GtkTextLayout",
	"__mbstate_t" => "VOID",

	# integer size 1
	"char" => "LONG",
	"signed char" => "char",
	"unsigned char" => "char",

	# integer size 2
	"short unsigned int" => "LONG",
	"short int" => "LONG",

	# integer size 4
	"int" => "LONG",
	"unsigned int" => "LONG",
	"long int" => "LONG",
	"long unsigned int" => "LONG",

	# integer size 8
	"long long int" => "LONGLONG",
	"long long unsigned int" => "LONGLONG",

	# other types
	"double" => "DOUBLE",
	"float" => "DOUBLE",
	"..." => "VARARG",
#	"__builtin_va_list" => "VALIST",
);

my $w;
for $w (keys %typedefs_override) {
	$typedefs{$w} = $typedefs_override{$w};
}

my $structs = 0;
my $attributes = 0;

# key = name, value = [ id, fields ]
my %structinfo;

# if $struct_name is "" --> substructure; return the last line.
sub parse_struct($$);
sub parse_struct($$) {
	my ($struct_type, $struct_name) = @_;

	my $fields = [];
	my ($name, $bitsize, $bitpos, $type);

	$structs ++;

	if ($struct_name ne "" and not $struct_name =~ /^_/) {
		print STDERR "  Warning: Structure name doesn't start with _: $struct_name\n";
	}

	while (<LIST>) {
		chop;
		last if $_ eq "};";
		$attributes ++;
		my $name = "?";
		my $bitsize = "?";

		# extract bitsize and bitpos
		if (/; \/\* bitsize 0x([0-9a-f]+), bitpos 0x([0-9a-f]+) \*\/$/) {
		 	$bitsize = hex($1);
			$bitpos = hex($2);
		}


		s/\/\*.*?\*\///g;	# remove comments
		s/ +/ /g;		# collapse whitespace
		s/^ //;			# leading space
		s/ $//;			# trailing space

		# end of substructure
		if (/^}\)? (.*?);$/) {
			$name = $1;
			$type = "END";
		}

		# two dimensional array
		elsif (/^(.*?)([a-zA-Z0-9_]+)\[0x([0-9a-f]+)\]\[0x([0-9a-f]+)\]:uint32:uint32;$/) {
			$type = $1;
		 	$name = $2;
		}

		# array of function pointers
		elsif (/^(.*?)\(\*([a-zA-Z0-9_]+)\)\[0x([0-9a-f]+)\]:uint32;$/) {
			$type = "GdkFilterFunc";		# XXX hack
		 	$name = $2;
		}

		# array
		elsif (/^(.*?)([a-zA-Z0-9_]+)\[0x([0-9a-f]+)\]:uint32;$/) {
			$type = $1;
		 	$name = $2;
		}

		# function pointer
		elsif (/^(.*?)\(\*([a-zA-Z0-9_]+)\) \(.*?\);$/) {
			$type = "func $1";	# was: "$1*"
		 	$name = $2;
		}
		
		# substructure
		elsif (/^\(?(union|struct) (%anon\d+) {/) {
			my $sub_fields;
			($name, $bitsize, $bitpos, $sub_fields) = 
				parse_struct($1, "");
			# print STDERR "result: $name $bitsize $bitpos $sub_fields\n";
			my $field;
			for $field (@{$sub_fields}) {
				# print STDERR "   + field $name.$$field[0]\n";
				push @$fields, [ "$name.$$field[0]",
					$$field[1],
					$$field[2] + $bitpos,
					$$field[3] ];
			}
			next;
		}

		# regular entry
		elsif (/^(.*?)([a-zA-Z0-9_]+);$/) {
			$type = $1;
			$name = $2;
			if ($type =~ /^struct \%anon\d+/) {
				print STDERR "WARNING: anonymous structure at line $.\n";
				$type = "long long int";	# XXX hack
			}
		}

		# unparseable line.
		else {
			print STDERR "$. unparseable struct entry: $_\n";
		}


		if ($type eq "END" and $struct_name eq "_") {
			$typedefs{"$struct_type $name"} = "STRUCT";
			unshift @$fields, -1; # $next_struct_nr;
			# $next_struct_nr ++;
			$structinfo{"$struct_type $name"} = $fields;
			return;
		}

		if ($name eq "?" or $bitsize eq "?") {
			print STDERR "$.    $name $bitsize: $_\n";
			next;
		}

		$type =~ s/ \*/*/g;		# space before *
		$type =~ s/ $//;		# remove trailing space

		if ($type eq "END") {
			return ($name, $bitsize, $bitpos, $fields);
		}

		# print STDERR "struct line >>$_<< --> '$name' >>$type<<\n";

		$c_types{$type} ++;
		push @$fields, [ $name, $bitsize, $bitpos, $type ];
	}

	if ($struct_name eq "") {
		print STDERR "$. ERROR -- reached this for substructure\n";
		return "";
	}


	$typedefs{"$struct_type $struct_name"} = "STRUCT";
	unshift @$fields, -1; # $next_struct_nr;
	# print STDERR "Structure $struct_name --> $#$fields\n";
	# $next_struct_nr ++;
	$structinfo{"$struct_type $struct_name"} = $fields;
	return "";
}

sub parse_enum($) {
	my @enumlist = split(/, /, $_[0]);
	my $s;
	my $i = 0;

	for $s (@enumlist) {
		if ($s =~ /^(\S+) = ([0-9a-fx]+)$/) {
			$s = $1;
			$i = $2;
			if ($i =~ /^0xffffffffffff/) {
				# negative!
				$i = substr($i, -4);
				$i = hex($i);
				$i = $i - 65536;
			} else {
				$i = hex($i);
			}
		}
		if (defined $enum{$s}) {
			print STDERR "  Warning: redefinition of enum $s\n";
		}
		$enum{$s} = $i;
		$i ++;
	}
}

# -- read debugging symbols to get a list of types --

my ($rv, $fn, $s);
open LIST, "objdump --debugging $tmpo|" or die;
while (<LIST>) {
	chop;

	next if $_ eq "";
	next if /^tmp\..*?\.(c|o):/;

	# path of source file
	if (/^ \/.*\.h:$/) {
		next;
	}

	
	s/\/\*.*?\*\///g;	# remove comments
	s/ +/ /g;		# collapse whitespace

	# def of a structure or union
	if (/^(struct|union) ([a-zA-Z0-9_]+) {/) {
		parse_struct($1, $2);
		next;
	}

	# structure or union with typedef
	if (/^typedef (struct|union) %anon\d+ {/) {
		parse_struct($1, "_");
		next;
	}

	# opaque structure or union
	if (/^typedef (struct|union) %anon\d+ (\S+);$/) {
		$typedefs{"$1 $2"} = "STRUCT";
		next;
	}

	# enum -- extract the constants and store them.
	if (/^enum (\S+) { ?([^}]+?) ?};$/) {
		$typedefs{"enum $1"} = "ENUM";
		parse_enum($2);
		next;
	}

	if (/^typedef enum { ?([^}]+?) ?} (.*);$/) {
		$typedefs{$2} = "ENUM";
		parse_enum($1);
		next;
	}

	# simple mapping of a structure or union
	if (/^typedef (struct|union) ([a-zA-Z0-9_]+) (\*)?([a-zA-Z0-9_]+)/) {
		my $ptr = defined $3 ? "*" : "";
		$typedefs{$4} = "$1 $2$ptr";
		#$typedefs{$4} = uc($1) . (defined $3 ? "PTR" : "");
		next;
	}

	# function
	if (/^typedef ([a-zA-Z_ *]+)\((\*)?([^)]+)\)/) {
		$rv = $1;		# return value
		$fn = $3;		# function name
		$rv =~ s/ $//;
		if (defined $2) {
			$rv .= "*";
		}
		# print STDERR "register a function $fn\n";
		$typedefs{$fn} = "FUNC";
		next;
	}

	# special case -- order seems to be reversed sometimes!!
	if (/^typedef ([a-zA-Z0-9_]+) (([a-zA-Z_ ]+) )?(int|char|double);/) {
		next if defined $typedefs{$1};
		$s = (defined $2 ? $2 : "") . $4;
		$typedefs{$1} = $s if $1 ne $s;
		next;
	}

	# normal case: typedef blah blah blah target;
	if (/^typedef ([a-zA-Z0-9_* ]+?)([a-zA-Z0-9_]+);/) {
		my $dest = $1;
		my $source = $2;
		$dest =~ s/ +$//;
		next if $source eq $dest;
		next if /complex/;
		if (defined $typedefs{$source}) {
			if (not defined $typedefs_override{$source}) {
				print STDERR "  Warning: redefinition of $source in line $.\n";
			}
			next;
		}
		$dest =~ s/ \*/*/;
#		print STDERR "typedef >>$source<< --> >>$dest<<\n";
		$typedefs{$source} = $dest;
		next;
	}

	# unrecognized entry!
	print STDERR "Cannot parse line $.: $_\n";
}
close LIST;
unlink $tmpo;

# -- DEBUG --
#for $w (sort keys %typedefs) {
#	print STDERR "typedefs: $w -> $typedefs{$w}\n";
#}
# -- END DEBUG --


# key = lua type, value = index of this type (for output)
my %luatypes;
my $nextluatype = '0';
# my $nextluatype_nodata = 'A';
# my $nextluatype_withdata = 'a';

sub make_luatype($) {
	my ($type) = @_;
	my $c;

	$c = $nextluatype;
	$nextluatype = chr(ord($nextluatype) + 1);
	# $nextluatype ++;

#	if (defined $with_detail{$type}) {
#		$c = $nextluatype_withdata;
#		$nextluatype_withdata = chr(ord($nextluatype_withdata) + 1);
#	} else {
#		$c = $nextluatype_nodata;
#		$nextluatype_nodata = chr(ord($nextluatype_nodata) + 1);
#	}

	return $c;
}

# %c_types is a list of data types as used in function arguments and in
# structures.

# %typedefs is a list of known typedefs, plus a few of my own to map everything
# to a few standard types that I can use in my code.

# The types in %c_types may have to be mapped multiple times via %typedefs
# to reach a final type.


# key = C type, value = Lua type
my %luatypemap;
my $mapped = 0;
my $mapped_weighed = 0;
my $total_weight = 0;

# Register a type.  Use the typedefs to find the basic type, e.g. an integer
# or a known structure.
sub add_type($) {
	my ($w) = @_;
	my $s = $w;
	my $ptr = "";
	my $weight;
	my $is_func = 0;

	if ($s =~ /^func (.*)/) {
		$s = $1;
		$is_func = 1;
	}

	# Try to find in typedefs; strip "*" (pointer) until found or
	# not a pointer anymore.  This allows to override "char*" for example.
	for (;;) {
		# simple mapping?
		if (defined $typedefs{$s}) {
			$s = $typedefs{$s};
			next;
		}

		# if "*" at end, remove it, try to map, and append "*" again.
		last unless substr($s, -1, 1) eq "*";
		$s = substr($s, 0, length($s) - 1);
		if (defined $typedefs{$s}) {
			$s = $typedefs{$s} . "*";
			next;
		}
		$ptr .= "PTR";
	}

	# not found?
	$s .= $ptr;
	if ($s eq $w) {
		print STDERR "unknown type >>$s<<\n";
	}

#	print STDERR "new luatype $w -> $s\n";

	# some special cases
	if ($s eq "") {
		print STDERR "Type reduced to nothing: $w\n";
	}
	if ($s eq "STRUCTPTR" and $w =~ /^(Gtk|Gdk)[A-Z]/) {
		$s = "WIDGET";
	}
	if ($s eq "STRUCTPTRPTR" and $w =~ /^(Gtk|Gdk)[A-Z]/) {
		$s = "WIDGETPTR";
	}

	if ($s ne $w) {
		if ($s !~ /^[A-Z]+$/) {
			print STDERR "Strange type $s\n";
			return;
		}
		if (not defined $luatypes{$s}) {
			# print STDERR "new type $nextluatype = $s\n";
			$luatypes{$s} = make_luatype($s); # $nextluatype;
			# inc_luatype();
		}
		$luatypemap{$w} = $s;
		$mapped ++;
	}
}

for $w (sort keys %c_types) {
	# maybe use a typedef that wasn't available at the time
	if ($w =~ /^struct _(.*)$/) {
		my $s = $1;
		my $s2 = $w;
		for (;;) {
			if (defined $typedefs{$s} and $typedefs{$s} eq $s2) {
				$w = $1;
				last;
			}
			last if substr($s, -1, 1) ne "*";
			$s = substr($s, 0, -1);
			$s2 = substr($s2, 0, -1);
		}
	}
	add_type($w);
}

# DEBUG
#for $w (sort keys %luatypemap) {
#	printf STDERR "lua %-30s %s\n", $w, $luatypemap{$w};
#}

# -- print a list of types (short) --
open OFILE, ">$ofile.types.c" or die;
# print OFILE "#include \"structinfo.h\"\n";
print OFILE "const struct ffi_type_map_t ffi_type_map[] = {\n";
my %tmp;
foreach $w (keys %luatypes) {
	$tmp{$luatypes{$w}} = $w;
}
foreach $w (sort keys %tmp) {
	my $name = $tmp{$w};
	my $ffi = $ffi_types{$name};
	if (not defined $ffi) {
		print STDERR "Basic type $name not known!!\n";
		exit(1);
	}
	printf OFILE "  { \"%s\", %d, %s, %s },\n",
		$name, $$ffi[0], $$ffi[1], defined $$ffi[2] ? "&ffi_type_$$ffi[2]" : "NULL";
#		$w=$tmp{$w}\n";
}
print OFILE "};\n";
close OFILE;


# A given type is (maybe) not available.  Try to create it using existing
# typedefs.
sub try_to_make_luatype($) {
	my ($t) = @_;

	return 1 if defined $luatypemap{$t};

	# print STDERR "try_to_make_luatype $t\n";
	my $s = $luatypemap{"$t*"};
	if (defined $s) {
		add_type($t);
		return 1;
	}
	print STDERR "ERROR: try_to_make_luatype for $t failed\n";
	return 0;
}

# print all enums
open OFILE, ">$ofile.enums" or die;
for $w (sort keys %enum) {
	my $val = $enum{$w};
	my $s = "";

	while ($val) {
		$s = (sprintf "\\%03o", $val & 0xff) . $s;
		$val = $val >> 8;
	}

	print OFILE "$w,$s\n";
}
close OFILE;


# -- find out which structures can be seen --

my %visible;

sub make_visible($);
sub make_visible($) {
	my ($arg) = @_;

	# print "make visible: $arg\n";

	# find basic type.  Don't care whether it's a pointer or not, the
	# result is the same -- that the structure declaration must be
	# included.
	my $ifo = $arg;
	$ifo =~ s/\*+$//;
	for (;;) {
		last if $ifo =~ /^(struct|union) (.*)$/;
		# print "   map $ifo\n";
		$ifo = $typedefs{$ifo};
		if (not defined $ifo) {
			# print STDERR "ERROR: $arg is not a structure after all\n";
			# print "   -> basic\n";
			$visible{$arg} = "BASIC";
			return;
		}

	}

	# print "   -> structure\n";

	my $s = $structinfo{$ifo};
	if (not defined $s) {
		if (substr($ifo, -1, 1) eq "*") {
			$s = $structinfo{substr($ifo, 0, -1)};
		}
		if (not defined $s) {
			print STDERR "  Error: referenced structure "
				. "$arg -> $ifo not known!\n";
			return;
		}
	}
	# both of these are required.
	$visible{$arg} = "YES";
	$visible{$ifo} = "YES";
	my $first = 1;
	my $attr;
	# print STDERR "- attributes of $arg\n";
	for $attr (@$s) {
		$first = 0, next if $first;
		my $s = $attr->[3];
		$s =~ s/^func //;
		make_visible($s) unless defined $visible{$s};
	}
}

for $func (keys %functions) {
	for $arg (@{$functions{$func}}) {
		my $s = $arg;
		$s =~ s/^func //;
		next if defined $visible{$s};
		# print "visible func $func -> arg $s\n";
		make_visible($s);
	}
}


# First, make a list of the mapped names.  This is required for proper sorting,
# as the number assignment must happen in sorted order.  Also, weed out
# unused structures.
my %structmap;
for $s (sort keys %structinfo) {

#	my $nr = $structinfo{$s}->[0];
#	next if $nr == -1;

#	if ($s !~ /^(union|struct) _(.*)/) {
#		print STDERR "  Warning: structure name doesn't start with _: $s\n";
#		next;
#	}

	# only emit visible structures
	if (not defined $visible{$s}) {
		# print STDERR "unused structure $s\n";
		next;
	}

	# go up one level in the type hierarchy -- should work all the time
	my $struct_name = $s;
	if ($s =~ /^(struct|union) _(.*)/ and defined $typedefs{$2} and $typedefs{$2} eq $s) {
		$struct_name = $2;
	} else {
		print STDERR "  Warning: no typedef for $s\n";
		next;
	}

	$structmap{$struct_name} = $s;
}


my $next_struct_nr = 0;

print STDERR "* assigning numbers to structures\n";

# print structure definitions
# XXX skip those that contain "---" as types (invalid types)
my $struct_count = 0;

# Assign structure numbers in sorted order.
my $s2;
for $s2 (sort keys %structmap) {
	$s = $structmap{$s2};
	# next unless defined $visible{$s};
	# my $ar = $structinfo{$s};
	$structinfo{$s}->[0] = $next_struct_nr;
	$next_struct_nr += 1;
	# print "structure $s -> " . $structinfo{$s}->[0] . "\n";
	$struct_count ++;
}

# returns: (type_code, type_detail)
sub to_luatype($) {
	my ($type) = @_;

	# sometimes a typedef for a structure pointer isn't used...
	if ($type =~ /^struct _(\S+)$/) {
		my $tmp = $1;
#		print STDERR "looking for typedef $tmp\n";
		for (;;) {
			if (defined $typedefs{$tmp} and $typedefs{$tmp} eq "struct _$tmp") {
				if (try_to_make_luatype($tmp)) {
					$type = $tmp;
				}
				last;
			}

			last if substr($tmp, -1, 1) ne "*";
			$tmp = substr($tmp, 0, -1);
		}
	}

	if (not defined $luatypemap{$type}) {
		print STDERR "ERROR: unknown lua type >>$type<<\n";
		return "---";
	}

	my $m = $luatypemap{$type};
	my $detail = -1;
	my $is_func = 0;

	# If this type has structure information, try to find it.
	if ($ffi_types{$m}->[0]) {
		$s = $type;
		$s =~ s/\*+$//;

		# find what structure that is
		my $ifo = $s;
		if ($ifo =~ /^func (.*)$/) {
			$ifo = $1;
			$is_func = 1;
			# print STDERR "/ func -> $ifo\n";
		}

		# use the type defs to find the basic type
		for (;;) {
			last if $ifo =~ /^(struct|union) (.*)$/;
			# print STDERR "/ $ifo -> $typedefs{$ifo}\n";
			$ifo = $typedefs{$ifo};
			if (not defined $ifo) {
				print STDERR "ERROR: $s is not a structure after all\n";
				return ($luatypes{$m}, $detail);
			}
		}

		# found?  if so, must be a structure or union.
		my $ent = $structinfo{$ifo};
		if (not defined $ent) {
			print STDERR "  Error: unknown structure $s ($ifo) in function\n";
		} elsif ($#$ent == -1) {
			print STDERR "  Error: in to_luatype, structure $ifo is empty\n";
		} elsif ($ent->[0] == -1) {
			print STDERR "  Error: in to_luatype, structure $ifo has no nr.\n";
		} else {
			$s = $ent->[0];	# replace with number
			$detail = $s;
		}
	}
	return ($luatypes{$m}, $detail);
}


# - print structures -

open OFILE, ">$ofile.structs.c" or die;
print OFILE "#include \"structinfo.h\"\n";
print OFILE "const struct struct_elem elem_list[] = {\n";

my (@out, @one);
my $elem_nr = 0;

print STDERR "* generating structures\n";

# In a shared library, each (absolute) pointer must be fixed up after loading.
# To avoid having one such fixup per string, I collect all of the strings
# in one large string and store offset instead of pointers.  Like GCC, I
# don't want same strings to be stored multiple times.

my %string_table;		# key = string, value = offset
my @string_table;		# array of strings so far
my $string_offset = 0;		# offset for next string

sub store_string($) {
	my ($s) = @_;
	my $ofs;

	# existing string?
	$ofs = $string_table{$s};
	return $ofs if defined $ofs;

	# enter a new one
	$ofs = $string_offset;
	$string_offset += length($s) + 1;	# plus NUL byte
	push @string_table, $s;
	$string_table{$s} = $ofs;

	return $ofs;
}

my $struct_name;
my $ofs;
for $struct_name (sort keys %structmap) {

	my $s = $structmap{$struct_name};

	# internal check
	my $nr = $structinfo{$s}->[0];
	if ($nr == -1) {
		print STDERR "INTERNAL ERROR: mark5\n";
		next;
	}

	my $elem_count = 0;
	my $attr;
	my $first = 1;
	my $struct_size = 0;
	@one = ();
	for $attr (@{$structinfo{$s}}) {
		$first=0, next if $first;
		my ($name, $bitsize, $bitpos, $type) = @{$attr};
		my ($type_code, $type_detail) = to_luatype($type);
		if ($type_code eq "---") {
			print STDERR "* ERROR: unknown data type $type\n";
			# fatal error: can't generate the structure, but
			# the numbers have been assigned and must be
			# contiguous.
			exit(1);
		}
		# next OUTER if $type_code eq "---";
		$ofs = store_string($$attr[0]);
		push @one, " {$ofs, $bitpos, $bitsize, $type_detail, '$type_code' },\n";
		$elem_count ++;
		if ($bitpos + $bitsize > $struct_size) {
			$struct_size = $bitpos + $bitsize;
		}
	}
	for $attr (@one) {
		print OFILE $attr;
	}

	$ofs = store_string($struct_name);
	$struct_size = int(($struct_size + 31) / 32) * 4;
	push @out, " { $ofs, $elem_nr, $elem_count, $struct_size }, /* $struct_name */\n";
	$elem_nr += $elem_count;
}
print OFILE "};\n\nconst struct struct_info struct_list[] = {\n";
for $s (@out) {
	print OFILE $s;
}
print OFILE "};\n";
printf OFILE "const int struct_count = %d;\n\n", $#out + 1;

# now emit the string table.
print OFILE "const char struct_strings[] =\n";
for $s (@string_table) {
	print OFILE " \"$s\\000\"\n";
}
print OFILE ";\n";

close OFILE;

# print function list with mapped arguments; the library wants to know the
# number of entries first, so generate & count entries, print the count and
# then all the entries.
print STDERR "* generating functions\n";
my $maxlen = 0;
my $maxargs = 0;
open OFILE, ">$ofile.funcs" or die;
for $func (sort keys %functions) {
	$args = $functions{$func};
	my @args = map {
		my ($type, $detail) = to_luatype($_);
		if ($detail == -1) {
			$type;
		} else {
			sprintf("%s\\%03o\\%03o", $type, $detail & 0xff,
				$detail >> 8);
		}
	} @$args;
	my $out = sprintf "%s,%s\n", $func, join('', @args);
	if ($out =~ /---/) {
		print STDERR "  Warning: skipping $out";
		next;
	}
	print OFILE $out;
	$maxlen = ($maxlen < length($out) ? length($out) : $maxlen);
	# $#args includes the return type, but numbering starts with 0, so...
	$maxargs = ($maxargs < $#args ? $#args : $maxargs);
}

close OFILE;


print STDERR "Max. line length is $maxlen, max. argument count $maxargs\n";
print STDERR "Total structures $structs, total attributes $attributes\n";



