
#include <string.h>
#include "hash.h"
extern unsigned int calc_hash(const char *p, int len);


const unsigned char *hash_search(const char *key, const struct hash_info *hi)
{
	unsigned int hash, index, i;
	unsigned const char *p;
	unsigned char c;
	int bytenr, count;

	/* calculate hash and get the index */
	hash = calc_hash(key, strlen(key));
	index = hash % hi->hash_size;

	/* get offset, which is stored as char, short or int */
	switch (hi->ofs_size) {
		case 1:
			index = hi->ofs[index];
			break;
		case 2:
			index = ((unsigned short*)hi->ofs)[index];
			break;
		case 4:
			index = ((unsigned int*)hi->ofs)[index];
			break;
	}

	if (!index)
		return NULL;

	/* index is 1-based in order to use 0 for empty buckets. */
	p = hi->data + index - 1;

	/* for full hash, compare the complete hash values */
	/* layout: bucket size, entries... */
	/* entry: hash value, data length, data bytes */
	if (hi->full_hash) {
		count = (*p++) + 1;
		for (i=count; i; i--) {
			if (*(unsigned int*)p == hash)
				goto found1;
			p += p[4] + 5;
		}
		return NULL;
found1:
		/* skip hash value */
		p += 4;
	} else {
		/* get the bucket size and the hash byte nr */
		c = *p++;
		bytenr = (c >> 6);
		count = (c & 0x3f) + 1;
		i = count;

		if (count > 1) {
			c = ((unsigned char*)&hash)[bytenr];
			for (; i; i--) {
				if (*p == c)
					goto found2;
				p += p[1] + 2;
			}
			return NULL;
found2:
			/* skip hash byte */
			p ++;
		}
	}

	/* p points to the data length, followed by data bytes.  count is the
	 * number of items in the bucket, whereas count-i is the number of the
	 * found entry.
	 */

	return p;
}
