/**
 * This is a minimal C program to start the Lua interpreter, load the standard
 * libraries, and run the file specified on the command line.  Error handling
 * is very good.  It is an experiment to save some memory, but it's not very
 * much compared to the standard lua executable, at least when considering the
 * memory used by Gtk2.
 * Lua-Gtk2 could be compiled in for no obvious benefit.
 *
 * by Wolfgang Oertl 2005
 */

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"

static const luaL_reg lualibs[] = {
  {"base", luaopen_base},
  {"table", luaopen_table},
  {"io", luaopen_io},
  {"string", luaopen_string},
  {"math", luaopen_math},
  {"debug", luaopen_debug},
  {"loadlib", luaopen_loadlib},
  /* add your libraries here */
/*  {"_gtk", luaopen_gtk2}, */
  {NULL, NULL}
};


int main(int argc, char **argv)
{
	lua_State *L;
	char *fname;
	int rc;

	if (argc != 2) {
		printf("Parameter: LUA file\n");
		return 1;
	}

	L = lua_open();

	/* open libraries */
	const luaL_reg *lib;
	for (lib = lualibs; lib->func; lib++) {
		lib->func(L);
		lua_settop(L, 0);
	}
	

	fname = argv[1];
	rc = luaL_loadfile(L, fname);
	if (rc) {
		printf("loadfile returned %d\n", rc);
	} else {
		rc = lua_pcall(L, 0, 0, 0);
		if (rc) {
			const char *s = lua_tostring(L, -1);
			printf("pcall returned %d: %s\n", rc, s);
		}
	}

	lua_close(L);
	return rc;
}

	


